@@ -2,7 +2,7 @@
 
   digest -- hash digest functions for R
 
-  Copyright (C) 2003 - 2016  Dirk Eddelbuettel <edd@debian.org>
+  Copyright (C) 2003 - 2022  Dirk Eddelbuettel <edd@debian.org>
 
   This file is part of digest.
 
@@ -36,6 +36,11 @@
 #include "zlib.h"
 #include "xxhash.h"
 #include "pmurhash.h"
+#include "blake3.h"
+
+#ifdef _WIN32
+#include <Windows.h>
+#endif
 
 unsigned long ZEXPORT digest_crc32(unsigned long crc,
                                    const unsigned char FAR *buf,
@@ -43,6 +48,27 @@ unsigned long ZEXPORT digest_crc32(unsigned long crc,
 
 static const char *sha2_hex_digits = "0123456789abcdef";
 
+FILE* open_with_widechar_on_windows(const char* txt) {
+    FILE* out;
+#ifdef _WIN32
+    wchar_t* buf;
+    size_t len = MultiByteToWideChar(CP_UTF8, 0, txt, -1, NULL, 0);
+    if (len <= 0) {
+        error("Cannot convert file to Unicode: %s", txt);
+    }
+    buf = (wchar_t*) R_alloc(len, sizeof(wchar_t));
+    if (buf == NULL) {
+        error("Could not allocate buffer of size: %ll", len);
+    }
+
+    MultiByteToWideChar(CP_UTF8, 0, txt, -1, buf, len);
+    out = _wfopen(buf, L"rb");
+#else
+    out = fopen(txt, "rb");
+#endif
+
+    return out;
+}
 
 SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Seed) {
     size_t BUF_SIZE = 1024;
@@ -55,7 +81,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
     int leaveRaw = INTEGER_VALUE(Leave_raw);
     SEXP result = R_NilValue;
     char output[128+1], *outputp = output;    /* 33 for md5, 41 for sha1, 65 for sha256, 128 for sha512; plus trailing NULL */
-    uint32_t nChar;
+    R_xlen_t nChar;
     int output_length = -1;
     if (IS_RAW(Txt)) { /* Txt is either RAW */
         txt = (char*) RAW(Txt);
@@ -67,6 +93,13 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
     } else { /* or a string */
         txt = (char*) STRING_VALUE(Txt);
         nChar = strlen(txt);
+
+        if (algo >= 100) {
+            fp = open_with_widechar_on_windows(txt);
+            if (!fp)  {
+              error("Cannot open input file: %s", txt);  /* #nocov */
+            }
+        }
     }
     if (skip > 0 && algo < 100) {
         if (skip>=nChar) {
@@ -91,7 +124,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
 
         if (!leaveRaw)
             for (j = 0; j < 16; j++)
-                sprintf(output + j * 2, "%02x", md5sum[j]);
+                snprintf(output + j * 2, 3, "%02x", md5sum[j]);
 
         break;
     }
@@ -108,7 +141,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
 
         if (!leaveRaw)
             for ( j = 0; j < 20; j++ )
-                sprintf( output + j * 2, "%02x", sha1sum[j] );
+                snprintf( output + j * 2, 3, "%02x", sha1sum[j] );
 
         break;
     }
@@ -119,7 +152,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         val  = digest_crc32(0L, 0, 0);
         val  = digest_crc32(val, (unsigned char*) txt, (unsigned) l);
 
-        sprintf(output, "%08x", (unsigned int) val);
+        snprintf(output, 128, "%08x", (unsigned int) val);
         break;
     }
     case 4: {     /* sha256 case */
@@ -135,7 +168,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
 
         if (!leaveRaw)
             for ( j = 0; j < 32; j++ )
-                sprintf( output + j * 2, "%02x", sha256sum[j] );
+                snprintf( output + j * 2, 3, "%02x", sha256sum[j] );
 
         break;
     }
@@ -165,21 +198,37 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
     }
     case 6: {     /* xxhash32 case */
         unsigned int val =  XXH32(txt, nChar, seed);
-        sprintf(output, "%08x", val);
+        snprintf(output, 128, "%08x", val);
         break;
     }
     case 7: {     /* xxhash64 case */
         unsigned long long val =  XXH64(txt, nChar, seed);
-#ifdef WIN32
-        sprintf(output, "%016" PRIx64, val);
+#if defined(WIN32) && !defined(_UCRT)
+        snprintf(output, 128, "%016" PRIx64, val);
 #else
-        sprintf(output, "%016llx", val);
+        snprintf(output, 128, "%016llx", val);
 #endif
         break;
     }
     case 8: {     /* MurmurHash3 32 */
         unsigned int val = PMurHash32(seed, txt, nChar);
-        sprintf(output, "%08x", val);
+        snprintf(output, 128, "%08x", val);
+        break;
+    }
+    case 10: {     /* blake3 */
+        output_length = BLAKE3_OUT_LEN;
+        blake3_hasher hasher;
+        blake3_hasher_init(&hasher);
+        blake3_hasher_update(&hasher, txt, nChar);
+        uint8_t val[BLAKE3_OUT_LEN];
+        blake3_hasher_finalize(&hasher, val, BLAKE3_OUT_LEN);
+        if (leaveRaw) {
+            memcpy(output, val, BLAKE3_OUT_LEN);
+        } else {
+            for (size_t i = 0; i < BLAKE3_OUT_LEN; i++) {
+                snprintf(output + i * 2, 3, "%02x", val[i]);
+            }
+        }
         break;
     }
     case 101: {     /* md5 file case */
@@ -189,9 +238,6 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         unsigned char buf[BUF_SIZE];
         unsigned char md5sum[16];
 
-        if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt); /* already covered at R level too */ /* #nocov */
-        }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         md5_starts( &ctx );
         if (length>=0) {
@@ -205,12 +251,11 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
             while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0)
                 md5_update( &ctx, buf, nChar );
         }
-        fclose(fp);
         md5_finish( &ctx, md5sum );
         memcpy(output, md5sum, 16);
         if (!leaveRaw)
             for (j = 0; j < 16; j++)
-                sprintf(output + j * 2, "%02x", md5sum[j]);
+                snprintf(output + j * 2, 3, "%02x", md5sum[j]);
         break;
     }
     case 102: {     /* sha1 file case */
@@ -220,9 +265,6 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         unsigned char buf[BUF_SIZE];
         unsigned char sha1sum[20];
 
-        if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt); /* already covered at R level too */ /* #nocov */
-        }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         sha1_starts ( &ctx );
         if (length>=0) {
@@ -235,21 +277,17 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
             while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0)
                 sha1_update( &ctx, buf, nChar );
         }
-        fclose(fp);
         sha1_finish ( &ctx, sha1sum );
         memcpy(output, sha1sum, 20);
         if (!leaveRaw)
             for ( j = 0; j < 20; j++ )
-                sprintf( output + j * 2, "%02x", sha1sum[j] );
+                snprintf( output + j * 2, 3, "%02x", sha1sum[j] );
         break;
     }
     case 103: {     /* crc32 file case */
         unsigned char buf[BUF_SIZE];
         unsigned long val;
 
-        if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt); /* already covered at R level too */ /* #nocov */
-        }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         val  = digest_crc32(0L, 0, 0);
         if (length>=0) {
@@ -262,8 +300,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
             while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0)
                 val  = digest_crc32(val , buf, (unsigned) nChar);
         }
-        fclose(fp);
-        sprintf(output, "%08x", (unsigned int) val);
+        snprintf(output, 128, "%08x", (unsigned int) val);
         break;
     }
     case 104: {     /* sha256 file case */
@@ -273,9 +310,6 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         unsigned char buf[BUF_SIZE];
         unsigned char sha256sum[32];
 
-        if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt); /* already covered at R level too */ /* #nocov */
-        }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         sha256_starts ( &ctx );
         if (length>=0) {
@@ -288,12 +322,11 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
             while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0)
                 sha256_update( &ctx, buf, nChar );
         }
-        fclose(fp);
         sha256_finish ( &ctx, sha256sum );
         memcpy(output, sha256sum, 32);
         if (!leaveRaw)
             for ( j = 0; j < 32; j++ )
-                sprintf( output + j * 2, "%02x", sha256sum[j] );
+                snprintf( output + j * 2, 3, "%02x", sha256sum[j] );
         break;
     }
     case 105: {     /* sha2-512 file case */
@@ -304,9 +337,6 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
 
         unsigned char buf[BUF_SIZE];
 
-        if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt); /* already covered at R level too */ /* #nocov */
-        }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         SHA512_Init(&ctx);
         if (length>=0) {
@@ -319,7 +349,6 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
             while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0)
                 SHA512_Update( &ctx, buf, nChar );
         }
-        fclose(fp);
 
 		/* Calling SHA512_Final, because SHA512_End will already
 		   convert the hash to a string, and we also want RAW */
@@ -342,9 +371,6 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         unsigned char buf[BUF_SIZE];
         XXH32_state_t* const state = XXH32_createState();
 
-        if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt); /* already covered at R level too */ /* #nocov */
-        }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         XXH_errorcode const resetResult = XXH32_reset(state, seed);
         if (resetResult == XXH_ERROR) {
@@ -367,20 +393,16 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
               }
             }
         }
-        fclose(fp);
         unsigned int val =  XXH32_digest(state);
         XXH32_freeState(state);
 
-        sprintf(output, "%08x", val);
+        snprintf(output, 128, "%08x", val);
         break;
     }
     case 107: {     /* xxhash64 */
         unsigned char buf[BUF_SIZE];
         XXH64_state_t* const state = XXH64_createState();
 
-        if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt); /* already covered at R level too */ /* #nocov */
-        }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         XXH_errorcode const resetResult = XXH64_reset(state, seed);
         if (resetResult == XXH_ERROR) {
@@ -403,14 +425,13 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
               }
             }
         }
-        fclose(fp);
         unsigned long long val =  XXH64_digest(state);
         XXH64_freeState(state);
 
 #ifdef WIN32
-        sprintf(output, "%016" PRIx64, val);
+        snprintf(output, 128, "%016" PRIx64, val);
 #else
-        sprintf(output, "%016llx", val);
+        snprintf(output, 128, "%016llx", val);
 #endif
         break;
     }
@@ -419,9 +440,6 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         unsigned char buf[BUF_SIZE];
         size_t total_length = 0;
 
-        if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt); /* already covered at R level too */ /* #nocov */
-        }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         if (length>=0) {
             while( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0
@@ -437,10 +455,37 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
                 total_length += nChar;
             }
         }
-        fclose(fp);
         unsigned int val = PMurHash32_Result(h1, carry, total_length);
 
-        sprintf(output, "%08x", val);
+        snprintf(output, 128, "%08x", val);
+        break;
+    }
+    case 110: {     /* blake3 file case */
+        output_length = BLAKE3_OUT_LEN;
+        unsigned char buf[BUF_SIZE];
+        uint8_t val[BLAKE3_OUT_LEN];
+        blake3_hasher hasher;
+        blake3_hasher_init(&hasher);
+
+        if (skip > 0) fseek(fp, skip, SEEK_SET);
+        if (length>=0) {
+            while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0 && length>0) {
+                if (nChar>length) nChar=length;
+                blake3_hasher_update( &hasher, buf, nChar );
+                length -= nChar;
+            }
+        } else {
+            while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0)
+                blake3_hasher_update( &hasher, buf, nChar );
+        }
+        blake3_hasher_finalize(&hasher, val, BLAKE3_OUT_LEN);
+        if (leaveRaw) {
+            memcpy(output, val, BLAKE3_OUT_LEN);
+        } else {
+            for (size_t i = 0; i < BLAKE3_OUT_LEN; i++) {
+                snprintf(output + i * 2, 3, "%02x", val[i]);
+            }
+        }
         break;
     }
     default: {
@@ -448,6 +493,10 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
     }
     } /* end switch */
 
+    if (algo >= 100 && fp) {
+        fclose(fp);
+    }
+
     if (leaveRaw && output_length > 0) {
         PROTECT(result=allocVector(RAWSXP, output_length));
         memcpy(RAW(result), output, output_length);
@@ -459,3 +508,25 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
 
   return result;
 }
+
+
+SEXP vdigest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Seed){
+    R_xlen_t n = length(Txt);
+    if (TYPEOF(Txt) == RAWSXP || n == 0)
+        return(digest(Txt, Algo, Length, Skip, Leave_raw, Seed));
+    SEXP ans = PROTECT(allocVector(STRSXP, n));
+    SEXP d = R_NilValue;
+    if (TYPEOF(Txt) == VECSXP){
+        for (R_xlen_t i = 0; i < n; i++){
+            d = digest(VECTOR_ELT(Txt, i), Algo, Length, Skip, Leave_raw, Seed);
+            SET_STRING_ELT(ans, i, STRING_ELT(d, 0));
+        }
+    } else {
+        for (R_xlen_t i = 0; i < n; i++){
+            d = digest(STRING_ELT(Txt, i), Algo, Length, Skip, Leave_raw, Seed);
+            SET_STRING_ELT(ans, i, STRING_ELT(d, 0));
+        }
+    }
+    UNPROTECT(1);
+    return ans;
+}
